
// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF MACHINE LEARNING RESEARCH (JMLR)
// AND JMLR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PUR- POSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JMLR, NOR JMLR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting
// the generality of the foregoing, neither the author, nor JMLR, nor
// JMLR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

// (C) 2011-2015 Urun Dogan, Tobias Glasmachers, Christian Igel.
// This program is part of the supplementary material of the JMLR
// paper 'A Unified View on Multi-class Support Vector Classification'.
// It requires version 3.0 of the Shark open source machine learning
// library. This program is free software; you can redistribute it
// and/or modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3, or
// (at your option) any later version.


#include <LinAlg/Base.h>
#include <Rng/GlobalRng.h>
#include <Data/Dataset.h>
#include <Data/CVDatasetTools.h>
#include <Data/Csv.h>
#include <Models/LinearModel.h>
#include <Models/Converter.h>
#include <Models/ConcatenatedModel.h>
#include <Models/Kernels/GaussianRbfKernel.h>
#include <ObjectiveFunctions/Loss/ZeroOneLoss.h>
#include <Algorithms/Trainers/McSvmMMRTrainer.h>
#include <Algorithms/Trainers/McSvmOVATrainer.h>
#include <Algorithms/Trainers/McSvmWWTrainer.h>
#include <Algorithms/Trainers/McSvmCSTrainer.h>
#include <Algorithms/Trainers/McSvmLLWTrainer.h>
#include <Algorithms/Trainers/McSvmADMTrainer.h>
#include <Algorithms/Trainers/McSvmATSTrainer.h>
#include <Algorithms/Trainers/McSvmATMTrainer.h>
#include <Algorithms/Trainers/McReinforcedSvmTrainer.h>
#include <Algorithms/Trainers/NormalizeComponentsUnitVariance.h>

#include "json.h"

#include <boost/format.hpp>
#include <iostream>
#include <sys/resource.h>

using namespace std;
using namespace shark;


void help()
{
	cout << endl
		<< "Multi Class Support Vector Machine demonstration program" << endl
		<< endl
		<< "(C) 2011-2015 Urun Dogan, Tobias Glasmachers, Christian Igel." << endl
		<< "This program is part of the supplementary material of the JMLR" << endl
		<< "paper 'A Unified View on Multi-class Support Vector Classification'." << endl
		<< "It requires version 3.0 of the Shark open source machine learning" << endl
		<< "library. This program is free software; you can redistribute it" << endl
		<< "and/or modify it under the terms of the GNU General Public License" << endl
		<< "as published by the Free Software Foundation; either version 3, or" << endl
		<< "(at your option) any later version." << endl
		<< endl
		<< "This program trains and evaluates a single multi-class SVM model." << endl
		<< "It can be called in two modes, namely cross-validation mode and" << endl
		<< "test mode:" << endl
		<< "./MultiClassSVM <logfile> <fold> <machine> <dataset> <C> <gamma> <seed>" << endl
		<< "./MultiClassSVM <logfile>  test  <machine> <dataset> <C> <gamma> <seed>" << endl
		<< "with parameters" << endl
		<< "    <logfile>    a json object holding accuracy and training time" << endl
		<< "    <fold>       5-fold cross-validation, values: 0, 1, 2, 3, 4" << endl
		<< "    <machine>    values: OVA, MMR, WW, CS, LLW, ADM, ATS, ATM, or RM" << endl
		<< "    <dataset>    data set file name, to which the postfixes '_train'" << endl
		<< "                 and '_test' and the extensions '.csv' and '.libsvm'" << endl
		<< "                 are appended. E.g., for files 'iris_train.csv' and" << endl
		<< "                 'iris_test.csv' provide the stem 'iris' here." << endl
		<< "    <C>          log2 of the regularization parameter C" << endl
		<< "    <gamma>      log2 of the kernel bandwidth parameter gamma" << endl
		<< "    <seed>       seed for the random number generator for creating" << endl
		<< "                 folds and training/test splits" << endl
		<< endl
		<< "The program trains the corresponding machine on the specified" << endl
		<< "training (sub)set and outputs its performance on the test or" << endl
		<< "validation set, the number of iterations, and the training time." << endl
		<< endl;
	exit(0);
}


int main(int argc, char** argv)
{
	try
	{
		setpriority(PRIO_PROCESS, 0, 20);

		// process command line parameters
		if (argc < 8) help();
		const char* logfilename = argv[1];
		bool testmode = (strcmp(argv[2], "test") == 0);
		unsigned int fold = strtol(argv[2], NULL, 10);
		const char* machine = argv[3];
		const char* dataset = argv[4];
		double C = strtod(argv[5], NULL);
		double gamma = strtod(argv[6], NULL);
		unsigned int seed = strtol(argv[7], NULL, 10);
		bool bias = false;

		C = pow(2.0, C);
		gamma = pow(2.0, gamma);

		// sanity checks
		if (! testmode && (fold > 4)) help();
		if (seed == 0) help();

		// create kernel and learning machine
		GaussianRbfKernel<> kernel(gamma);
		AbstractSvmTrainer<RealVector, unsigned int>* trainer = NULL;
		if (strcmp(machine, "OVA") == 0) trainer = new McSvmOVATrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "MMR") == 0) trainer = new McSvmMMRTrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "WW") == 0) trainer = new McSvmWWTrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "CS") == 0) trainer = new McSvmCSTrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "LLW") == 0) trainer = new McSvmLLWTrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "ADM") == 0) trainer = new McSvmADMTrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "ATS") == 0) trainer = new McSvmATSTrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "ATM") == 0) trainer = new McSvmATMTrainer<RealVector>(&kernel, C, bias);
		else if (strcmp(machine, "RM") == 0) trainer = new McReinforcedSvmTrainer<RealVector>(&kernel, C, bias);
		else help();

		// read and normalize data
		ClassificationDataset train, test;
		boost::format fmterTrain("data/%s_train.csv");
		boost::format fmterTest("data/%s_test.csv");
		if (testmode)
		{
			// load data from two files, one for training, one for testing
			try
			{
				importCSV(train, (fmterTrain % dataset).str(), LAST_COLUMN, ',', '#');
			} catch (...) {
				cerr << "cannot open data file " << (fmterTrain % dataset).str() << " for reading" << endl;
				exit(EXIT_FAILURE);
			}
			try
			{
				importCSV(test , (fmterTest  % dataset).str(), LAST_COLUMN, ',', '#');
			} catch (...) {
				cerr << "cannot open data file " << (fmterTest  % dataset).str() << " for reading" << endl;
				exit(EXIT_FAILURE);
			}

			// shuffle training and test sets
			size_t n = train.numberOfElements();
			train.append(test);
			Rng::seed(seed);
			train.shuffle();
			test = splitAtElement(train, n);

			// normalize data based on the training set
			unsigned int dim = inputDimension(train);
			Normalizer<RealVector> normalizer(dim, true);
			NormalizeComponentsUnitVariance<RealVector> normalizationTrainer(true);
			normalizationTrainer.train(normalizer, train.inputs());
			train = transformInputs(train, normalizer);
			test = transformInputs(test, normalizer);
		}
		else
		{
			// load only training data and split them into training and
			// validation according to the five-fold cross-validation scheme
			ClassificationDataset tmp;
			try {
				importCSV(tmp, (fmterTrain % dataset).str(), LAST_COLUMN, ',', '#');
			} catch (...) {
				cerr << "cannot open data file " << (fmterTrain % dataset).str() << " for reading" << endl;
				exit(EXIT_FAILURE);
			}
			unsigned int dim = inputDimension(tmp);
			Normalizer<RealVector> normalizer(dim, true);
			NormalizeComponentsUnitVariance<RealVector> normalizationTrainer(true);
			normalizationTrainer.train(normalizer, tmp.inputs());
			tmp = transformInputs(tmp, normalizer);
			Rng::seed(seed);
			CVFolds<ClassificationDataset> folds = createCVSameSize(tmp, 5);
			train = folds.training(fold);
			test = folds.validation(fold);
		}

		unsigned int classes = std::max(numberOfClasses(train), numberOfClasses(test));

		if (testmode) cout << "mode: test" << endl;
		else cout << "mode: cross-validation; fold: " << fold << " (seed: " << seed << ")" << endl;
		cout << "machine: " << machine << endl
			<< "    C: " << C << endl
			<< "gamma: " << gamma << endl
			<< endl;

		// training
		cout << "machine training (this can take a while) ..." << flush;
		KernelClassifier<RealVector> svm;
		trainer->stoppingCondition().maxIterations = 10000 * classes * train.numberOfElements();
		trainer->shrinking() = false;
		trainer->train(svm, train);
		cout << " done." << endl;

		// testing
		cout << "evaluating ..." << flush;
		ZeroOneLoss<unsigned int> loss;
		Data<unsigned int> output = svm(test.inputs());
		double accuracy = 100.0 * (1.0 - loss.eval(test.labels(), output));
		cout << " accuracy: " << accuracy << " %" << endl;

		// training process summary
		cout << endl << "training summary:" << endl;
		cout << "#iterations: " << trainer->solutionProperties().iterations << endl;
		cout << "       time: " << trainer->solutionProperties().seconds << " seconds" << endl;

		// write log file
		Json json(object);
		json["accuracy"] = accuracy;
		json["time"] = trainer->solutionProperties().seconds;
		json["C"] = C;
		json["gamma"] = gamma;
		json["fold"] = (double)fold;
		json["seed"] = (double)seed;
		json["dataset"] = dataset;
		json["machine"] = machine;
		json["sv"] = (double)svm.decisionFunction().basis().numberOfElements();
		json.save(logfilename, true);
	}
	catch (shark::Exception const& ex)
	{
		cout << endl << endl << "EXCEPTION: " << ex.what() << endl << endl;
	}
}
