
// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF MACHINE LEARNING RESEARCH (JMLR)
// AND JMLR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PUR- POSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT. THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JMLR, NOR JMLR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE. Without limiting
// the generality of the foregoing, neither the author, nor JMLR, nor
// JMLR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

// (C) 2012-2015 Urun Dogan, Tobias Glasmachers, Christian Igel.
// This program is part of the supplementary material of the JMLR
// paper 'A Unified View on Multi-class Support Vector Classification'.
// It requires version 3.0 of the Shark open source machine learning
// library. This program is free software; you can redistribute it
// and/or modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3, or
// (at your option) any later version.


#include <LinAlg/Base.h>
#include <Rng/GlobalRng.h>
#include <Data/Dataset.h>
#include <Data/CVDatasetTools.h>
#include <Data/Csv.h>
#include <Data/Libsvm.h>
#include <Models/LinearModel.h>
#include <Models/Normalizer.h>
#include <ObjectiveFunctions/Loss/ZeroOneLoss.h>
#include <Algorithms/Trainers/McSvmMMRTrainer.h>
#include <Algorithms/Trainers/McSvmOVATrainer.h>
#include <Algorithms/Trainers/McSvmWWTrainer.h>
#include <Algorithms/Trainers/McSvmCSTrainer.h>
#include <Algorithms/Trainers/McSvmLLWTrainer.h>
#include <Algorithms/Trainers/McSvmATSTrainer.h>
#include <Algorithms/Trainers/McSvmADMTrainer.h>
#include <Algorithms/Trainers/McSvmATMTrainer.h>
#include <Algorithms/Trainers/McReinforcedSvmTrainer.h>
#include <Algorithms/Trainers/NormalizeComponentsUnitVariance.h>

#include "json.h"

#include <boost/format.hpp>
#include <iostream>
#include <sys/resource.h>

using namespace std;
using namespace shark;


void help()
{
	cout << endl
		<< "Linear Multi Class Support Vector Machine demonstration program" << endl
		<< endl
		<< "(C) 2012-2015 Urun Dogan, Tobias Glasmachers, Christian Igel." << endl
		<< "This program is part of the supplementary material of the JMLR" << endl
		<< "paper 'A Unified View on Multi-class Support Vector Classification'." << endl
		<< "It requires version 3.0 of the Shark open source machine learning" << endl
		<< "library. This program is free software; you can redistribute it" << endl
		<< "and/or modify it under the terms of the GNU General Public License" << endl
		<< "as published by the Free Software Foundation; either version 3, or" << endl
		<< "(at your option) any later version." << endl
		<< endl
		<< "This program trains and evaluates a single multi-class SVM model." << endl
		<< "It can be called in two modes, namely cross-validation mode and" << endl
		<< "test mode:" << endl
		<< "./MultiClassSVMLinear <logfile> <fold> <machine> <dataset> <C> <seed>" << endl
		<< "./MultiClassSVMLinear <logfile>  test  <machine> <dataset> <C> <seed>" << endl
		<< "with parameters" << endl
		<< "    <logfile>    a json object holding accuracy and training time" << endl
		<< "    <fold>       5-fold cross-validation, values: 0, 1, 2, 3, 4" << endl
		<< "    <machine>    values: OVA, MMR, WW, CS, LLW, ADM, ATS, ATM, or RM" << endl
		<< "    <dataset>    data set file name, to which the postfixes '_train'" << endl
		<< "                 and '_test' and the extensions '.csv' and '.libsvm'" << endl
		<< "                 are appended. E.g., for files 'iris_train.csv' and" << endl
		<< "                 'iris_test.csv' provide the stem 'iris' here." << endl
		<< "    <C>          log2 of the regularization parameter C" << endl
		<< "    <seed>       seed for the random number generator for creating" << endl
		<< "                 folds and training/test splits" << endl
		<< "    normalize    optional flag; normalize features to unit variance" << endl
		<< endl
		<< "The program trains the corresponding machine on the specified" << endl
		<< "training (sub)set and outputs its performance on the test or" << endl
		<< "validation set, the number of iterations, and the training time." << endl
		<< endl;
	exit(0);
}


LabeledData<CompressedRealVector, unsigned int> loaddata(string dataset, bool train)
{
	try
	{
		LabeledData<CompressedRealVector, unsigned int> data;
		string filename = "data/" + dataset + (train ? "_train" : "_test") + ".libsvm";
		import_libsvm(data, filename.c_str());
		return data;
	}
	catch (...)
	{
	}

	try
	{
		LabeledData<RealVector, unsigned int> rawdata;
		string filename = "data/" + dataset + (train ? "_train" : "_test") + ".csv";
		importCSV(rawdata, filename, LAST_COLUMN);
		LabeledData<CompressedRealVector, unsigned int> data(rawdata.numberOfBatches());
		for (size_t i=0; i<rawdata.numberOfBatches(); i++)
		{
			data.inputs().batch(i) = rawdata.inputs().batch(i);
			data.labels().batch(i) = rawdata.labels().batch(i);
		}
		return data;
	}
	catch (...)
	{
	}

	cout << endl << endl << "FATAL ERROR: failed to load data set: " << dataset << endl << endl;
	exit(1);
}

int main(int argc, char** argv)
{
	try
	{
		setpriority(PRIO_PROCESS, 0, 20);

		// process command line parameters
		if (argc < 7) help();
		const char* logfilename = argv[1];
		bool testmode = (strcmp(argv[2], "test") == 0);
		unsigned int fold = strtol(argv[2], NULL, 10);
		const char* machine = argv[3];
		const char* dataset = argv[4];
		double C = strtod(argv[5], NULL);
		unsigned int seed = strtol(argv[6], NULL, 10);
		bool normalize = (argc > 7 && strcmp(argv[7], "normalize") == 0);

		C = pow(2.0, C);

		// sanity checks
		if (! testmode && (fold > 4)) help();
		if (seed == 0) help();

		// create kernel and learning machine
		AbstractLinearSvmTrainer<CompressedRealVector>* trainer = NULL;
		if (strcmp(machine, "OVA") == 0) trainer = new LinearMcSvmOVATrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "MMR") == 0) trainer = new LinearMcSvmMMRTrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "WW") == 0) trainer = new LinearMcSvmWWTrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "CS") == 0) trainer = new LinearMcSvmCSTrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "LLW") == 0) trainer = new LinearMcSvmLLWTrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "ADM") == 0) trainer = new LinearMcSvmADMTrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "ATS") == 0) trainer = new LinearMcSvmATSTrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "ATM") == 0) trainer = new LinearMcSvmATMTrainer<CompressedRealVector>(C);
		else if (strcmp(machine, "RM") == 0) trainer = new LinearMcSvmReinforcedTrainer<CompressedRealVector>(C);
		else help();

		// dump configuration
		cout << "CONFIGURATION:" << endl;
		cout << "logfile ......... " << logfilename << endl;
		cout << "mode ............ ";
		if (testmode) cout << "test" << endl;
		else cout << "fold " << fold << endl;
		cout << "machine ......... " << machine << endl;
		cout << "dataset ......... " << dataset << endl;
		cout << "C ............... " << C << endl;
		cout << "seed ............ " << seed << endl;
		cout << "normalize ....... " << (normalize ? "true" : "false") << endl;
		cout << endl;

		// check whether the logfile already exists
		try
		{
			Json j;
			if (j.load(logfilename)) exit(0);   // log file exists
		}
		catch (...)
		{ }

		// read and normalize data
		LabeledData<CompressedRealVector, unsigned int> train, test;
		if (testmode)
		{
			train = loaddata(dataset, true);
			test = loaddata(dataset, false);

			// shuffle training and test sets
			size_t n = train.numberOfElements();
			train.append(test);
			Rng::seed(seed);
			train.shuffle();
			test = splitAtElement(train, n);

			if (normalize)
			{
				// normalize data based on the training set
				unsigned int dim = dataDimension(train.inputs());
				Normalizer<CompressedRealVector> normalizer(dim, false);
				NormalizeComponentsUnitVariance<CompressedRealVector> normalizationTrainer(false);
				normalizationTrainer.train(normalizer, train.inputs());
				train = transformInputs(train, normalizer);
				test = transformInputs(test, normalizer);
			}
		}
		else
		{
			// load only training data and split them into training and
			// validation according to the five-fold cross-validation scheme
			LabeledData<CompressedRealVector, unsigned int> tmp = loaddata(dataset, true);

			if (normalize)
			{
				// normalize data
				unsigned int dim = tmp.inputs().element(0).size();
				Normalizer<CompressedRealVector> normalizer(dim, false);
				NormalizeComponentsUnitVariance<CompressedRealVector> normalizationTrainer(false);
				normalizationTrainer.train(normalizer, tmp.inputs());
				tmp = transformInputs(tmp, normalizer);
			}

			Rng::seed(seed);
			CVFolds< LabeledData<CompressedRealVector, unsigned int> > folds = createCVSameSize(tmp, 5);
			train = folds.training(fold);
			test = folds.validation(fold);
		}

		// determine number of classes (= maximal label + 1)
		unsigned int classes = std::max(numberOfClasses(train), numberOfClasses(test));

		if (testmode) cout << "mode: test" << endl;
		else cout << "mode: cross-validation; fold: " << fold << " (seed: " << seed << ")" << endl;
		cout << "machine: " << machine << endl
			<< "    C: " << C << endl
			<< endl;

		// training
		cout << "machine training (this can take a while) ..." << flush;
		LinearClassifier<CompressedRealVector> svm;
		trainer->stoppingCondition().maxIterations = 100000 * classes * train.numberOfElements();
		trainer->train(svm, train);
		cout << " done." << endl;

		// testing
		cout << "evaluating ..." << flush;
		ZeroOneLoss<unsigned int> loss;
		Data<unsigned int> output = svm(test.inputs());
		double accuracy = 100.0 * (1.0 - loss.eval(test.labels(), output));
		cout << " accuracy: " << accuracy << " %" << endl;

		// training process summary
		cout << endl << "training summary:" << endl;
		cout << "#iterations: " << trainer->solutionProperties().iterations << endl;
		cout << "       time: " << trainer->solutionProperties().seconds << " seconds" << endl;

		// write log file
		Json json(object);
		json["accuracy"] = accuracy;
		json["time"] = trainer->solutionProperties().seconds;
		json["iterations"] = (double)trainer->solutionProperties().iterations;
		json["C"] = C;
		json["fold"] = (double)fold;
		json["seed"] = (double)seed;
		json["dataset"] = dataset;
		json["machine"] = machine;
		json.save(logfilename, true);
	}
	catch (shark::Exception const& ex)
	{
		cout << endl << endl << "EXCEPTION: " << ex.what() << endl << endl;
	}
}
